import sys
import logging

logger = logging.getLogger("LiteAI-SDK")
logger.addHandler(logging.NullHandler())

def enable_logging(level=logging.INFO):
    """
    Enable logging for the LiteAI SDK.

    Args:
        level: The logging level (default: logging.INFO).

               Common values: logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR
    """
    logger.setLevel(level)

    if not any(isinstance(h, logging.StreamHandler) for h in logger.handlers):
        handler = logging.StreamHandler(sys.stderr)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)



import inspect
from typing import Any, Callable, TypeVar
from .toolset import Toolset
from ...types.tool import ToolLike

F = TypeVar("F", bound=Callable[..., Any])
TOOL_FLAG = "__is_tool__"

def python_tool(func: F) -> F:
    setattr(func, TOOL_FLAG, True)
    return func

class PythonToolset(Toolset):
    def get_tools(self) -> list[ToolLike]:
        return [
            method
            for _, method in inspect.getmembers(self, predicate=inspect.ismethod)
            if getattr(method, TOOL_FLAG, False)
        ]



from abc import ABC
from mcp.types import TextContent, ImageContent, AudioContent, ResourceLink, EmbeddedResource, TextResourceContents, BlobResourceContents
from ..._mcp_client.mcp_client import McpClient, Tool, ToolResult
from ..._mcp_client.local_mcp_client import LocalMcpClient, LocalServerParams
from ..._mcp_client.remote_mcp_client import RemoteMcpClient, RemoteServerParams, OAuthParams
from .toolset import Toolset
from ...types.tool import ToolDef, ToolLike

class McpToolset(ABC, Toolset):
    def __init__(self, client: McpClient):
        self._client = client
        self._tools_cache: list[ToolLike] | None = None

    def _mcp_tool_to_tool_def(self, mcp_tool: Tool) -> ToolDef:
        async def wrapper(**kwargs) -> str:
            result = await self._client.call_tool(mcp_tool.name, kwargs)
            return self._format_tool_result(result)

        tool_def = ToolDef(
            name=mcp_tool.name,
            description=mcp_tool.description or f"MCP tool: {mcp_tool.name}",
            parameters=mcp_tool.inputSchema,
            execute=wrapper
        )
        return tool_def

    def _format_tool_result(self, result: ToolResult) -> str:
        is_error, content = result.is_error, result.content
        content_parts = []

        if is_error:
            content_parts.append("Error executing tool:")

        for block in content:
            match block:
                case TextContent():
                    content_parts.append(block.text)
                case ImageContent():
                    content_parts.append(f"[Generated Image: {block.mimeType}]")
                case AudioContent():
                    content_parts.append(f"[Generated Audio: {block.mimeType}]")
                case ResourceLink():
                    details = [f"Resource Reference: {block.uri}"]
                    if block.mimeType: details.append(f"Type: {block.mimeType}")
                    if block.size: details.append(f"Size: {block.size} bytes")
                    if block.description: details.append(f"Description: {block.description}")
                    content_parts.append("\n".join(details))
                case EmbeddedResource():
                    resource = block.resource
                    header = f"Resource ({resource.uri}):"
                    if isinstance(resource, TextResourceContents):
                        content_parts.append(f"{header}\n{resource.text}")
                    elif isinstance(resource, BlobResourceContents):
                        content_parts.append(f"{header}\n[Binary data: {resource.mimeType}]")
                case _:
                    logger.warning(f"Unknown tool result block type: {type(block)}")

        return "\n\n".join(content_parts)

    async def connect(self) -> None:
        await self._client.connect()
        await self.refresh_tools()

    async def disconnect(self) -> None:
        await self._client.disconnect()
        self._tools_cache = None

    async def refresh_tools(self) -> None:
        mcp_tools = await self._client.list_tools()
        self._tools_cache = [self._mcp_tool_to_tool_def(tool) for tool in mcp_tools]

    def get_tools(self) -> list[ToolLike]:
        if self._tools_cache is None:
            raise RuntimeError(f"Not connected to MCP server. Call await {self.__class__.__name__}(...).connect() first")
        return self._tools_cache


class LocalMcpToolset(McpToolset):    
    def __init__(self, name: str, params: LocalServerParams):
        client = LocalMcpClient(name, params)
        super().__init__(client)

class RemoteMcpToolset(McpToolset):    
    def __init__(self, name: str, params: RemoteServerParams):
        client = RemoteMcpClient(name, params)
        super().__init__(client)



from __future__ import annotations

import json
import dataclasses
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Literal, NamedTuple, cast
from pydantic import BaseModel, ConfigDict, PrivateAttr, field_validator
from litellm.types.utils import (
    Message as LiteLlmMessage,
    ModelResponse as LiteLlmModelResponse,
    ModelResponseStream as LiteLlmModelResponseStream,
    Choices as LiteLlmModelResponseChoices,
    ChatCompletionAudioResponse,
    ChatCompletionMessageToolCall,
    ChatCompletionDeltaToolCall,
    Usage as LiteLlmUsage
)
from litellm.types.llms.openai import (
    AllMessageValues,
    OpenAIMessageContent,
    ChatCompletionAssistantToolCall,
    ImageURLListItem as ChatCompletionImageURL,

    ChatCompletionUserMessage,
    ChatCompletionAssistantMessage,
    ChatCompletionToolMessage,
    ChatCompletionSystemMessage,
)
from ..types.tool import ToolLike
from ..tool.utils import find_tool_by_name
from ..logger import logger

if TYPE_CHECKING:
    from . import LlmRequestParams

class ChatMessage(BaseModel, ABC):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
    )

    @abstractmethod
    def to_litellm_message(self) -> AllMessageValues: ...

class UserMessage(ChatMessage):
    content: OpenAIMessageContent
    role: Literal["user"] = "user"

    def to_litellm_message(self) -> ChatCompletionUserMessage:
        return ChatCompletionUserMessage(role=self.role, content=self.content)

class ToolMessage(ChatMessage):
    """
    The `tool_def` field is ref to the target tool of the tool call, and
    it will only be None when the target tool is not found
    """
    id: str
    name: str
    arguments: str
    result: str | None = None
    error: str | None = None
    role: Literal["tool"] = "tool"

    _tool_def: ToolLike | None = PrivateAttr(default=None)

    @field_validator("result", mode="before")
    def validate_result(cls, v: Any) -> Any:
        if v is None: return v
        if isinstance(v, str): return v
        return json.dumps(v, ensure_ascii=False)

    @property
    def tool_def(self) -> ToolLike | None:
        return self._tool_def

    def with_tool_def(self, tool_def: ToolLike) -> "ToolMessage":
        self._tool_def = tool_def
        return self

    def to_litellm_message(self) -> ChatCompletionToolMessage:
        if self.result is None and self.error is None:
            raise ValueError(f"ToolMessage({self.id}, {self.name}) is incomplete, "
                              "result and error cannot be both None")

        if self.error is not None:
            content = json.dumps({"error": self.error}, ensure_ascii=False)
        else:
            assert self.result is not None
            content = self.result

        return ChatCompletionToolMessage(
            role=self.role,
            content=content,
            tool_call_id=self.id)

class ToolCallTuple(NamedTuple):
    id: str
    function_name: str
    function_arguments: str

class AssistantMessage(ChatMessage):
    content: str | None = None
    reasoning_content: str | None = None
    tool_calls: list[ChatCompletionAssistantToolCall] | None = None
    audio: ChatCompletionAudioResponse | None = None
    images: list[ChatCompletionImageURL] | None = None
    usage: LiteLlmUsage | None = None
    role: Literal["assistant"] = "assistant"

    _request_params_ref: LlmRequestParams | None = PrivateAttr(default=None)

    @classmethod
    def from_litellm_message(cls, response: LiteLlmModelResponse) -> "AssistantMessage":
        choices = cast(list[LiteLlmModelResponseChoices], response.choices)
        message = choices[0].message

        tool_calls: list[ChatCompletionAssistantToolCall] | None = None
        if (message_tool_calls := message.get("tool_calls")) is not None:
            tool_calls = [ChatCompletionAssistantToolCall(
                id=tool_call.id,
                function={
                    "name": tool_call.function.name,
                    "arguments": tool_call.function.arguments,
                },
                type="function",
            ) for tool_call in cast(list[ChatCompletionMessageToolCall], message_tool_calls)]

        return cls.model_construct(
            content=message.get("content"),
            reasoning_content=message.get("reasoning_content"),
            tool_calls=tool_calls,
            audio=message.get("audio"),
            images=message.get("images"),
            usage=response.get("usage"),
        )

    def with_request_params(self, request_params: LlmRequestParams) -> "AssistantMessage":
        self._request_params_ref = request_params
        return self

    def to_litellm_message(self) -> ChatCompletionAssistantMessage:
        return ChatCompletionAssistantMessage(role=self.role,
                                              content=self.content or "",
                                              reasoning_content=self.reasoning_content,
                                              tool_calls=self.tool_calls)

    def parse_tool_calls(self) -> list[ToolCallTuple] | None:
        if self.tool_calls is None: return None
        results = []
        for tool_call in self.tool_calls:
            id = tool_call.get("id")
            function = tool_call.get("function") # this can not be None
            function_name = function.get("name")
            function_arguments = function.get("arguments")
            if id is None or\
               function is None or\
               function_name is None or\
               function_arguments is None:
                return None
            results.append(ToolCallTuple(id, function_name, function_arguments))
        return results

    def get_partial_tool_messages(self) -> list[ToolMessage] | None:
        """
        Get a partial tool message from the assistant message.
        The returned tool message is not complete,
        it only contains the tool call id, name and arguments.
        Returns None if there is no tool call in the assistant message.
        """
        has_tool_def = self._request_params_ref is not None and\
                       self._request_params_ref.tools is not None
        if not has_tool_def:
            logger.warning("AssistantMessage.get_partial_tool_messages() called without request params. "
                           "Call with_request_params() first to enable auto tool_def attachment feature.")

        parsed_tool_calls = self.parse_tool_calls()
        if parsed_tool_calls is None: return None

        results = []
        for tool_call in parsed_tool_calls:
            tool_message = ToolMessage(
                id=tool_call.id,
                name=tool_call.function_name,
                arguments=tool_call.function_arguments,
                result=None,
                error=None)

            if has_tool_def:
                assert self._request_params_ref and self._request_params_ref.tools
                target_tool = find_tool_by_name(self._request_params_ref.tools, tool_call.function_name)
                if target_tool:
                    tool_message = tool_message.with_tool_def(target_tool)
                else:
                    logger.warning(f"Tool {tool_call.function_name} not found in request params, "
                                    "tool_def will not be attached to the tool message")

            results.append(tool_message)
        return results

class SystemMessage(ChatMessage):
    content: str
    role: Literal["system"] = "system"

    def to_litellm_message(self) -> ChatCompletionSystemMessage:
        return ChatCompletionSystemMessage(role=self.role, content=self.content)

@dataclasses.dataclass
class TextChunk:
    content: str

@dataclasses.dataclass
class UsageChunk:
    input_tokens: int
    output_tokens: int
    total_tokens: int

@dataclasses.dataclass
class ReasoningChunk:
    content: str

@dataclasses.dataclass
class AudioChunk:
    data: ChatCompletionAudioResponse

@dataclasses.dataclass
class ImageChunk:
    data: list[ChatCompletionImageURL]

@dataclasses.dataclass
class ToolCallChunk:
    id: str | None
    name: str | None
    arguments: str
    index: int

MessageChunk = TextChunk | UsageChunk | ReasoningChunk | AudioChunk | ImageChunk | ToolCallChunk

def openai_chunk_normalizer(
        chunk: LiteLlmModelResponseStream
        ) -> list[MessageChunk]:
    if len(chunk.choices) == 0: return []

    result = []
    delta = chunk.choices[0].delta
    if delta.get("content"):
        result.append(TextChunk(cast(str, delta.content)))
    if delta.get("reasoning_content"):
        result.append(ReasoningChunk(cast(str, delta.reasoning_content)))
    if delta.get("audio"):
        result.append(AudioChunk(cast(ChatCompletionAudioResponse, delta.audio)))
    if delta.get("images"):
        result.append(ImageChunk(cast(list[ChatCompletionImageURL], delta.images)))
    if delta.get("tool_calls"):
        for tool_call in cast(list[ChatCompletionDeltaToolCall], delta.tool_calls):
            result.append(ToolCallChunk(
                tool_call.id,
                tool_call.function.name,
                tool_call.function.arguments,
                tool_call.index))
    if (usage := getattr(chunk, "usage", None)) is not None:
        usage = cast(LiteLlmUsage, usage)
        result.append(UsageChunk(
            input_tokens=usage.prompt_tokens,
            output_tokens=usage.completion_tokens,
            total_tokens=usage.total_tokens))
    return result



import dataclasses
from collections.abc import Callable
from typing import Any, Awaitable

ToolFn = Callable[..., Any] | Callable[..., Awaitable[Any]]

"""
RawToolDef example:
{
    "name": "get_current_weather",
    "description": "Get the current weather in a given location",
    "parameters": {
        "type": "object",
        "properties": {
            "location": {
                "type": "string",
                "description": "The city and state, e.g. San Francisco, CA",
            },
            "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
        },
        "required": ["location"],
    }
}
"""
RawToolDef = dict[str, Any]

@dataclasses.dataclass
class ToolDef:
    name: str
    description: str
    execute: ToolFn
    parameters: dict[str, Any] | None = None

ToolLike = ToolDef | RawToolDef | ToolFn



from typing import Any
from litellm.types.utils import LlmProviders
from .tool.prepare import prepare_tools
from .types import LlmRequestParams, ToolMessage
from .types.tool import ToolLike

ParsedParams = dict[str, Any]

class ParamParser:
    def __init__(self,
                 provider: LlmProviders,
                 base_url: str,
                 api_key: str):
        self._provider = provider
        self._base_url = base_url
        self._api_key = api_key

    @staticmethod
    def _extract_tool_params(params: LlmRequestParams) -> list[ToolLike] | None:
        if params.tools is None and params.toolsets is None:
            return None
        tools = []
        if params.tools:
            tools = params.tools
        if params.toolsets:
            for toolset in params.toolsets:
                tools.extend(toolset.get_tools())
        return tools

    def _parse(self, params: LlmRequestParams) -> ParsedParams:
        extracted_tool_likes = self._extract_tool_params(params)
        tools = extracted_tool_likes and prepare_tools(extracted_tool_likes)

        transformed_messages = []
        for message in params.messages:
            if type(message) is ToolMessage and\
               message.result is None and\
               message.error is None:
                # skip ToolMessage that is not resolved
                continue
            transformed_messages.append(message.to_litellm_message())

        return {
            "model": f"{self._provider.value}/{params.model}",
            "messages": transformed_messages,
            "base_url": self._base_url,
            "api_key": self._api_key,
            "tools": tools,
            "tool_choice": params.tool_choice,
            "timeout": params.timeout_sec,
            "extra_headers": params.headers,
            **(params.extra_args or {})
        }

    def parse_nonstream(self, params: LlmRequestParams) -> ParsedParams:
        parsed = self._parse(params)
        parsed["stream"] = False
        return parsed

    def parse_stream(self, params: LlmRequestParams) -> ParsedParams:
        parsed = self._parse(params)
        parsed["stream"] = True
        parsed["stream_options"] = {"include_usage": True}
        return parsed



import asyncio
import dataclasses
import queue
from typing import Any, Literal
from collections.abc import AsyncGenerator, Generator

from .tool import ToolLike
from ..tool.toolset import Toolset
from .message import ChatMessage, AssistantMessage, ToolMessage, MessageChunk

@dataclasses.dataclass
class LlmRequestParams:
    model: str
    messages: list[ChatMessage]
    tools: list[ToolLike] | None = None
    toolsets: list[Toolset] | None = None
    tool_choice: Literal["auto", "required", "none"] = "auto"
    execute_tools: bool = False

    timeout_sec: float | None = None
    temperature: float | None = None
    max_tokens: int | None = None
    headers: dict[str, str] | None = None

    extra_args: dict[str, Any] | None = None

# --- --- --- --- --- ---

GenerateTextResponse = list[AssistantMessage | ToolMessage]
StreamTextResponseSync = tuple[
    Generator[MessageChunk],
    queue.Queue[AssistantMessage | ToolMessage | None]]
StreamTextResponseAsync = tuple[
    AsyncGenerator[MessageChunk],
    asyncio.Queue[AssistantMessage | ToolMessage | None]]



import httpx
import webbrowser
from dataclasses import dataclass, field
from contextlib import AsyncExitStack
from typing import Any, NamedTuple
from mcp import ClientSession
from mcp.client.auth import OAuthClientProvider
from mcp.client.streamable_http import streamable_http_client
from mcp.shared.auth import OAuthClientMetadata
from pydantic import AnyUrl
from .oauth_server import LocalOAuthServer, OAuthCode, TokenStorage, InMemoryTokenStorage
from .mcp_client import McpClient, Tool, ToolResult, McpSessionNotEstablishedError
from ..logger import logger

@dataclass
class OAuthParams:
    oauth_scopes: list[str] | None = None
    oauth_timeout: int = 120
    oauth_token_storage: TokenStorage = field(default_factory= InMemoryTokenStorage)
@dataclass
class RemoteServerParams:
    url: str
    bearer_token: str | None = None
    oauth_params: OAuthParams | None = None
    http_headers: dict[str, str] | None = None

# --- --- --- --- --- ---

class OAuthContext(NamedTuple):
    client: httpx.AsyncClient
    server: LocalOAuthServer

class RemoteMcpClient(McpClient):
    def __init__(self, name: str, params: RemoteServerParams):
        self._name = name
        self._params = params
        self._session: ClientSession | None = None
        self._exit_stack: AsyncExitStack | None = None
        self._oauth_context: OAuthContext | None = self._init_oauth()

    def _init_http_headers(self) -> dict[str, str] | None:
        if self._params.http_headers is None and self._params.bearer_token is None:
            return None
        headers = {}
        if self._params.http_headers is not None:
            headers.update(self._params.http_headers)
        if self._params.bearer_token is not None:
            headers["Authorization"] = f"Bearer {self._params.bearer_token}"
        return headers

    def _init_oauth(self) -> OAuthContext | None:
        if self._params.oauth_params is None:
            return None
        
        oauth_params = self._params.oauth_params

        server = LocalOAuthServer(timeout=oauth_params.oauth_timeout)
        scopes = None
        if oauth_params.oauth_scopes is not None:
            scopes = " ".join(oauth_params.oauth_scopes)

        client_provider = OAuthClientProvider(
            server_url=self._params.url,
            client_metadata=OAuthClientMetadata(
                client_name=self._name,
                redirect_uris=[AnyUrl(server.callback_url)],
                grant_types=["authorization_code", "refresh_token"],
                response_types=["code"],
                scope=scopes,
                token_endpoint_auth_method="none",
            ),
            storage=oauth_params.oauth_token_storage,
            redirect_handler=self._handle_redirect,
            callback_handler=self._handle_oauth_callback,
        )
        client = httpx.AsyncClient(auth=client_provider,
                                   headers=self._init_http_headers(),
                                   follow_redirects=True)
        return OAuthContext(client, server)

    async def _handle_redirect(self, url: str) -> None:
        logger.info("[OAuth] Authentication required, opening browser...")
        logger.info(f"[OAuth] If browser does not open automatically, copy and open the following link: \n{url}\n")
        try:
            webbrowser.open(url)
        except Exception as e:
            logger.error(f"[OAuth] Not able to open browser: {e}")

    async def _handle_oauth_callback(self) -> OAuthCode:
        if self._oauth_context is None:
            raise ValueError("OAuth context not initialized")
        return await self._oauth_context.server.wait_for_code()

    async def connect(self):
        self._exit_stack = AsyncExitStack()
        if self._oauth_context:
            http_client = self._oauth_context.client
            await self._oauth_context.server.start()
        else:
            http_client = await self._exit_stack.enter_async_context(
                httpx.AsyncClient(headers=self._init_http_headers(), follow_redirects=True))

        try:
            read_stream, write_stream, _ = await self._exit_stack.enter_async_context(
                streamable_http_client(self._params.url, http_client=http_client)
            )
            self._session = await self._exit_stack.enter_async_context(
                ClientSession(read_stream, write_stream)
            )
            await self._session.initialize()
        except Exception:
            await self.disconnect()
            raise

    async def list_tools(self) -> list[Tool]:
        if not self._session:
            raise McpSessionNotEstablishedError()

        result = await self._session.list_tools()
        return result.tools

    async def call_tool(
        self, tool_name: str, arguments: dict[str, Any] | None = None
    ) -> ToolResult:
        if not self._session:
            raise McpSessionNotEstablishedError()

        response = await self._session.call_tool(tool_name, arguments)
        return ToolResult(response.isError, response.content)

    async def disconnect(self):
        try:
            if self._exit_stack:
                await self._exit_stack.aclose()
        finally:
            self._session = None
            self._exit_stack = None

            if self._oauth_context:
                try: await self._oauth_context.client.aclose()
                except Exception: pass
                try: await self._oauth_context.server.stop()
                except Exception: pass



import asyncio
import socket
import uvicorn
from starlette.applications import Starlette
from starlette.responses import HTMLResponse
from starlette.routing import Route
from starlette.requests import Request
from mcp.client.auth import TokenStorage
from mcp.shared.auth import OAuthClientInformationFull, OAuthToken

CALLBACK_PAGE = """
<html>
    <body style="font-family: sans-serif; text-align: center; padding-top: 50px;">
        <h2 style="color: green;">Login Successful!</h2>
        <p>You can close this window now.</p>
        <script>window.close();</script>
    </body>
</html>
"""

OAuthCode = tuple[str, str | None]
OAuthCallbackFuture = asyncio.Future[OAuthCode]

class InMemoryTokenStorage(TokenStorage):
    def __init__(self):
        self.tokens: OAuthToken | None = None
        self.client_info: OAuthClientInformationFull | None = None

    async def get_tokens(self) -> OAuthToken | None:
        return self.tokens

    async def set_tokens(self, tokens: OAuthToken) -> None:
        self.tokens = tokens

    async def get_client_info(self) -> OAuthClientInformationFull | None:
        return self.client_info

    async def set_client_info(self, client_info: OAuthClientInformationFull) -> None:
        self.client_info = client_info

def _find_free_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("0.0.0.0", 0))
        port = s.getsockname()[1]
        return port

class LocalOAuthServer:
    def __init__(self, timeout: int):
        self._port = _find_free_port()
        self._timeout = timeout
        self._future = OAuthCallbackFuture()
        self._server: uvicorn.Server | None = None
        self._server_task: asyncio.Task | None = None

    @property
    def callback_url(self) -> str:
        return f"http://localhost:{self._port}/callback"

    async def _handle_callback(self, request: Request):
        params = request.query_params

        # check if error
        if "error" in params:
            error = params.get("error")
            desc = params.get("error_description", "Unknown error")
            if not self._future.done():
                self._future.set_exception(RuntimeError(f"OAuth Error: {error} - {desc}"))
            return HTMLResponse(f"<h3>Auth failed: {error}</h3>", status_code=400)

        code = params.get("code")
        state = params.get("state")

        if not code:
            return HTMLResponse("<h3>Missing 'code' parameter</h3>", status_code=400)

        if not self._future.done():
            self._future.set_result((code, state))

        return HTMLResponse(CALLBACK_PAGE)

    async def wait_for_code(self) -> OAuthCode:
        auth_code =  await asyncio.wait_for(self._future, timeout=self._timeout)
        self._future = OAuthCallbackFuture()
        return auth_code

    async def start(self):
        routes = [
            Route("/callback", self._handle_callback, methods=["GET"])
        ]
        app = Starlette(routes=routes)

        config = uvicorn.Config(app=app, host="127.0.0.1", port=self._port, log_level="error")
        self._server = uvicorn.Server(config)
        self._server_task = asyncio.create_task(self._server.serve())

    async def stop(self):
        if self._server:
            self._server.should_exit = True
        if self._server_task:
            await self._server_task



from .toolset import Toolset
from .python_toolset import PythonToolset, python_tool
from .mcp_toolset import (
    McpToolset,
    LocalMcpToolset,
    RemoteMcpToolset,
    LocalServerParams,
    RemoteServerParams,
    OAuthParams,
)

__all__ = [
    "Toolset",

    "PythonToolset",
    "python_tool",

    "McpToolset",
    "LocalMcpToolset",
    "RemoteMcpToolset",
    "LocalServerParams",
    "RemoteServerParams",
    "OAuthParams",
]



from ..types.tool import ToolFn, ToolDef, ToolLike

def find_tool_by_name(tools: list[ToolLike], name: str) -> ToolLike | None:
    for tool in tools:
        if callable(tool) and tool.__name__ == name:
            return tool
        elif isinstance(tool, ToolDef) and tool.name == name:
            return tool
        elif isinstance(tool, dict) and tool.get("name") == name:
            return tool
    return None



import asyncio
import queue
from typing import cast
from collections.abc import AsyncGenerator, Generator
from litellm import CustomStreamWrapper, completion, acompletion
from litellm.exceptions import (
    AuthenticationError,
    PermissionDeniedError,
    RateLimitError,
    ContextWindowExceededError,
    BadRequestError,
    InvalidRequestError,
    InternalServerError,
    ServiceUnavailableError,
    ContentPolicyViolationError,
    APIError,
    Timeout,
)
from litellm.utils import get_valid_models
from litellm.types.utils import (
    LlmProviders,
    ModelResponse as LiteLlmModelResponse,
    ModelResponseStream as LiteLlmModelResponseStream
)
from .debug import enable_debugging
from .param_parser import ParamParser
from .stream import AssistantMessageCollector
from .tool.execute import execute_tool_sync, execute_tool
from .tool.toolset import (
    Toolset,
    python_tool, PythonToolset,
    McpToolset,
    LocalMcpToolset, LocalServerParams,
    RemoteMcpToolset, RemoteServerParams, OAuthParams,
)
from .tool.utils import find_tool_by_name
from .types import LlmRequestParams, GenerateTextResponse, StreamTextResponseSync, StreamTextResponseAsync
from .types.tool import ToolFn, ToolDef, RawToolDef, ToolLike
from .types.exceptions import *
from .types.message import (
    ChatMessage, UserMessage, SystemMessage, AssistantMessage, ToolMessage,
    MessageChunk, TextChunk, UsageChunk, ReasoningChunk, AudioChunk, ImageChunk, ToolCallChunk,
    ToolCallTuple, openai_chunk_normalizer
)
from .logger import logger, enable_logging

class LLM:
    """
    The `generate_text` and `stream_text` API will return ToolMessage in the returned sequence
    only if `params.execute_tools` is True.

    - - -

    Possible exceptions raises for `generate_text` and `stream_text`:
        - AuthenticationError
        - PermissionDeniedError
        - RateLimitError
        - ContextWindowExceededError
        - BadRequestError
        - InvalidRequestError
        - InternalServerError
        - ServiceUnavailableError
        - ContentPolicyViolationError
        - APIError
        - Timeout
    """

    def __init__(self,
                 provider: LlmProviders,
                 base_url: str,
                 api_key: str):
        self.provider = provider
        self.base_url = base_url
        self.api_key = api_key
        self._param_parser = ParamParser(self.provider, self.base_url, self.api_key)

    @staticmethod
    def _should_resolve_tool_calls(
            params: LlmRequestParams,
            message: AssistantMessage,
            ) -> tuple[list[ToolLike],
                       list[ToolCallTuple]] | None:
        parsed_tool_calls = message.parse_tool_calls()
        condition = params.execute_tools and\
                    params.tools is not None and\
                    parsed_tool_calls is not None
        if condition:
            assert params.tools is not None
            assert parsed_tool_calls is not None
            return params.tools, parsed_tool_calls
        return None

    @staticmethod
    async def _execute_tool_calls(
        tools: list[ToolLike],
        tool_call_tuples: list[ToolCallTuple]
        ) -> list[ToolMessage]:
        results = []
        for tool_call_tuple in tool_call_tuples:
            function_name = tool_call_tuple.function_name
            function_arguments = tool_call_tuple.function_arguments

            if (target_tool := find_tool_by_name(tools, function_name)) is None:
                logger.warning(f"Tool \"{function_name}\" not found, skipping execution.")
                continue
            if isinstance(target_tool, dict):
                logger.warning(f"Tool \"{function_name}\" is a raw tool, skipping execution.")
                continue

            result, error = None, None
            try:
                result = await execute_tool(target_tool, function_arguments)
            except Exception as e:
                error = f"{type(e).__name__}: {str(e)}"
            results.append(ToolMessage(
                id=tool_call_tuple.id,
                name=function_name,
                arguments=function_arguments,
                result=result,
                error=error).with_tool_def(target_tool))
        return results

    @staticmethod
    def _execute_tool_calls_sync(
        tools: list[ToolLike],
        tool_call_tuples: list[ToolCallTuple]
        ) -> list[ToolMessage]:
        results = []
        for tool_call_tuple in tool_call_tuples:
            function_name = tool_call_tuple.function_name
            function_arguments = tool_call_tuple.function_arguments

            if (target_tool := find_tool_by_name(tools, function_name)) is None:
                logger.warning(f"Tool \"{function_name}\" not found, skipping execution.")
                continue
            if isinstance(target_tool, dict):
                logger.warning(f"Tool \"{function_name}\" is a raw tool, skipping execution.")
                continue

            result, error = None, None
            try:
                result = execute_tool_sync(target_tool, function_arguments)
            except Exception as e:
                error = f"{type(e).__name__}: {str(e)}"
            results.append(ToolMessage(
                id=tool_call_tuple.id,
                name=function_name,
                arguments=function_arguments,
                result=result,
                error=error).with_tool_def(target_tool))
        return results

    def list_models(self) -> list[str]:
        return get_valid_models(
            custom_llm_provider=self.provider.value,
            check_provider_endpoint=True,
            api_base=self.base_url,
            api_key=self.api_key)

    def generate_text_sync(self, params: LlmRequestParams) -> GenerateTextResponse:
        response = completion(**self._param_parser.parse_nonstream(params))
        response = cast(LiteLlmModelResponse, response)
        assistant_message = AssistantMessage\
                            .from_litellm_message(response)\
                            .with_request_params(params)
        result: GenerateTextResponse = [assistant_message]
        if (tools_and_tool_calls := self._should_resolve_tool_calls(params, assistant_message)):
            tools, tool_calls = tools_and_tool_calls
            result += self._execute_tool_calls_sync(tools, tool_calls)
        return result

    async def generate_text(self, params: LlmRequestParams) -> GenerateTextResponse:
        response = await acompletion(**self._param_parser.parse_nonstream(params))
        response = cast(LiteLlmModelResponse, response)
        assistant_message = AssistantMessage\
                            .from_litellm_message(response)\
                            .with_request_params(params)
        result: GenerateTextResponse = [assistant_message]
        if (tools_and_tool_calls := self._should_resolve_tool_calls(params, assistant_message)):
            tools, tool_calls = tools_and_tool_calls
            result += await self._execute_tool_calls(tools, tool_calls)
        return result

    def stream_text_sync(self, params: LlmRequestParams) -> StreamTextResponseSync:
        def stream(response: CustomStreamWrapper) -> Generator[MessageChunk]:
            nonlocal message_collector
            for chunk in response:
                chunk = cast(LiteLlmModelResponseStream, chunk)
                yield from openai_chunk_normalizer(chunk)
                message_collector.collect(chunk)

            message = message_collector.get_message().with_request_params(params)
            full_message_queue.put(message)
            if (tools_and_tool_calls := self._should_resolve_tool_calls(params, message)):
                tools, tool_calls = tools_and_tool_calls
                tool_messages = self._execute_tool_calls_sync(tools, tool_calls)
                for tool_message in tool_messages:
                    full_message_queue.put(tool_message)
            full_message_queue.put(None)

        response = completion(**self._param_parser.parse_stream(params))
        message_collector = AssistantMessageCollector()
        returned_stream = stream(cast(CustomStreamWrapper, response))
        full_message_queue = queue.Queue[AssistantMessage | ToolMessage | None]()
        return returned_stream, full_message_queue

    async def stream_text(self, params: LlmRequestParams) -> StreamTextResponseAsync:
        async def stream(response: CustomStreamWrapper) -> AsyncGenerator[MessageChunk]:
            nonlocal message_collector
            async for chunk in response:
                chunk = cast(LiteLlmModelResponseStream, chunk)
                for normalized_chunk in openai_chunk_normalizer(chunk):
                    yield normalized_chunk
                message_collector.collect(chunk)

            message = message_collector.get_message().with_request_params(params)
            await full_message_queue.put(message)
            if (tools_and_tool_calls := self._should_resolve_tool_calls(params, message)):
                tools, tool_calls = tools_and_tool_calls
                tool_messages = await self._execute_tool_calls(tools, tool_calls)
                for tool_message in tool_messages:
                    await full_message_queue.put(tool_message)
            await full_message_queue.put(None)

        response = await acompletion(**self._param_parser.parse_stream(params))
        message_collector = AssistantMessageCollector()
        returned_stream = stream(cast(CustomStreamWrapper, response))
        full_message_queue = asyncio.Queue[AssistantMessage | ToolMessage | None]()
        return returned_stream, full_message_queue

__all__ = [
    # Exceptions
    "AuthenticationError",
    "PermissionDeniedError",
    "RateLimitError",
    "ContextWindowExceededError",
    "BadRequestError",
    "InvalidRequestError",
    "InternalServerError",
    "ServiceUnavailableError",
    "ContentPolicyViolationError",
    "APIError",
    "Timeout",

    "LLM",
    "LlmProviders",
    "LlmRequestParams",

    "Toolset",
    "python_tool",
    "PythonToolset",
    "McpToolset",
    "LocalMcpToolset",
    "RemoteMcpToolset",
    "LocalServerParams",
    "RemoteServerParams",
    "OAuthParams",

    "ToolFn",
    "ToolDef",
    "RawToolDef",
    "ToolLike",
    "execute_tool",
    "execute_tool_sync",

    "ChatMessage",
    "UserMessage",
    "SystemMessage",
    "AssistantMessage",
    "ToolMessage",

    "MessageChunk",
    "TextChunk",
    "UsageChunk",
    "ReasoningChunk",
    "AudioChunk",
    "ImageChunk",
    "ToolCallChunk",

    "GenerateTextResponse",
    "StreamTextResponseSync",
    "StreamTextResponseAsync",

    "enable_debugging",
    "enable_logging",
]



from litellm.exceptions import (  
    AuthenticationError,
    PermissionDeniedError,
    RateLimitError,
    ContextWindowExceededError,
    BadRequestError,
    InvalidRequestError,
    InternalServerError,
    ServiceUnavailableError,
    ContentPolicyViolationError,
    APIError,
    Timeout,
)

__all__ = [
    "AuthenticationError",
    "PermissionDeniedError",
    "RateLimitError",
    "ContextWindowExceededError",
    "BadRequestError",
    "InvalidRequestError",
    "InternalServerError",
    "ServiceUnavailableError",
    "ContentPolicyViolationError",
    "APIError",
    "Timeout",
]



import asyncio
import json
from functools import singledispatch
from typing import Any, Awaitable, Callable, cast
from types import FunctionType, MethodType, CoroutineType
from ..types.tool import ToolDef

async def _coroutine_wrapper(awaitable: Awaitable[Any]) -> CoroutineType:
    return await awaitable

def _arguments_normalizer(arguments: str | dict) -> dict:
    if isinstance(arguments, str):
        parsed = json.loads(arguments)
        return cast(dict, parsed)
    elif isinstance(arguments, dict):
        return arguments
    else:
        raise ValueError(f"Invalid arguments type: {type(arguments)}")

def _result_normalizer(result: Any) -> str:
    if isinstance(result, str):
        return result
    return json.dumps(result, ensure_ascii=False)

@singledispatch
def execute_tool_sync(tool, arguments: str | dict) -> str:
    raise ValueError(f"Invalid tool type: {type(tool)}")

@execute_tool_sync.register(FunctionType)
@execute_tool_sync.register(MethodType)
def _(toolfn: Callable, arguments: str | dict) -> str:
    arguments = _arguments_normalizer(arguments)
    result = asyncio.run(_coroutine_wrapper(toolfn(**arguments)))\
             if asyncio.iscoroutinefunction(toolfn)\
             else toolfn(**arguments)
    return _result_normalizer(result)

@execute_tool_sync.register(ToolDef)
def _(tooldef: ToolDef, arguments: str | dict) -> str:
    arguments = _arguments_normalizer(arguments)
    result = asyncio.run(_coroutine_wrapper(tooldef.execute(**arguments)))\
             if asyncio.iscoroutinefunction(tooldef.execute)\
             else tooldef.execute(**arguments)
    return _result_normalizer(result)

@singledispatch
async def execute_tool(tool, arguments: str | dict) -> str:
    raise ValueError(f"Invalid tool type: {type(tool)}")

@execute_tool.register(FunctionType)
@execute_tool.register(MethodType)
async def _(toolfn: Callable, arguments: str | dict) -> str:
    arguments = _arguments_normalizer(arguments)
    result = await toolfn(**arguments)\
             if asyncio.iscoroutinefunction(toolfn)\
             else toolfn(**arguments)
    return _result_normalizer(result)

@execute_tool.register(ToolDef)
async def _(tooldef: ToolDef, arguments: str | dict) -> str:
    arguments = _arguments_normalizer(arguments)
    result = await tooldef.execute(**arguments)\
             if asyncio.iscoroutinefunction(tooldef.execute)\
             else tooldef.execute(**arguments)
    return _result_normalizer(result)



import dataclasses
from litellm import ChatCompletionAssistantToolCall
from litellm.types.utils import ChatCompletionDeltaToolCall,\
                                ModelResponseStream as LiteLlmModelResponseStream
from .types.message import AssistantMessage

@dataclasses.dataclass
class ToolCallTemp:
    id: str | None = None
    name: str = ""
    arguments: str = ""

class ToolCallCollector:
    def __init__(self):
        self.tool_call_map: dict[int, ToolCallTemp] = {}

    def collect(self, tool_call_chunk: ChatCompletionDeltaToolCall):
        if tool_call_chunk.index not in self.tool_call_map:
            self.tool_call_map[tool_call_chunk.index] = ToolCallTemp()

        temp_tool_call = self.tool_call_map[tool_call_chunk.index]
        if tool_call_chunk.get("id"):
            temp_tool_call.id = tool_call_chunk.id
        if tool_call_chunk.function.get("name"):
            assert tool_call_chunk.function.name is not None
            temp_tool_call.name += tool_call_chunk.function.name
        if tool_call_chunk.function.get("arguments"):
            assert tool_call_chunk.function.arguments is not None
            temp_tool_call.arguments += tool_call_chunk.function.arguments

    def get_tool_calls(self) -> list[ChatCompletionAssistantToolCall]:
        return [{
            "id": tool_call.id,
            "function": {
                "name": tool_call.name,
                "arguments": tool_call.arguments,
            },
            "type": "function"
        } for tool_call in self.tool_call_map.values()]

class AssistantMessageCollector:
    def __init__(self):
        self.message_buf = AssistantMessage(content=None)
        self.tool_call_collector = ToolCallCollector()

    def collect(self, chunk: LiteLlmModelResponseStream):
        delta = chunk.choices[0].delta
        if delta.get("content"):
            assert delta.content is not None
            if self.message_buf.content is None:
                self.message_buf.content = ""
            self.message_buf.content += delta.content

        if delta.get("reasoning_content"):
            assert delta.reasoning_content is not None
            if self.message_buf.reasoning_content is None:
                self.message_buf.reasoning_content = ""
            self.message_buf.reasoning_content += delta.reasoning_content

        if delta.get("tool_calls"):
            assert delta.tool_calls is not None
            for tool_call_chunk in delta.tool_calls:
                self.tool_call_collector.collect(tool_call_chunk)

        if delta.get("images"):
            assert delta.images is not None
            if self.message_buf.images is None:
                self.message_buf.images = []
            self.message_buf.images = delta.images
        
        if delta.get("audio"):
            assert delta.audio is not None
            self.message_buf.audio = delta.audio

    def get_message(self) -> AssistantMessage:
        self.message_buf.tool_calls = self.tool_call_collector.get_tool_calls()
        return self.message_buf

    def clear(self):
        """
        This class will be created for each stream, so we don't need to clear it.
        """
